#include "gen.hpp"

#include <array>
#include <string>
#include <vector>
#include <algorithm>
#include <cassert>

using namespace std;

enum Tp {
    Os = 0,
    R,
    Ex
};

string tp_tostr[] = {"Os", "R", "Ex"};

struct Train {
    int arrival;
    Tp type;
};

struct Subtask {
    char const *name;
    int max_trains;
    int max_arrival = 1e9;
    int max_travel_times[3] = {100'000, 100'000, 100'000};
    bool no_Ex = false;
    bool no_overlap = false;
};

struct Problem {
    int n;
    int travel_times[3];
    vector<Train> trains;

    friend std::ostream &operator<<(std::ostream &os, const Problem &p) {
        os << p.n << '\n';
        os << p.travel_times[Os] << ' ' << p.travel_times[R] << ' ' << p.travel_times[Ex] << '\n';
        for(auto [arr, tp] : p.trains) {
            os << arr << ' ' << tp_tostr[tp] << '\n';
        }

        return os;
    }
};

struct Options {
    int min_spacing;
    int max_spacing;
    double zero_space_chance = 0;
    int min_travel_time[3];
    int max_travel_time[3];
    bool type_weights_rand = true;
    double type_weights[3];
    bool exact_n = false;
};

Problem gen(rng r, Subtask s, Options o) {
    uniform_real_distribution<double> prob_dist(0, 1);
    if(o.type_weights_rand) {
        for(int i = 0; i < 3; ++i) {
            o.type_weights[i] = prob_dist(r);
        }
    }
    if(s.no_Ex) o.type_weights[Ex] = 0;

    int n = s.max_trains;
    if(!o.exact_n) n *= (9 + prob_dist(r))/10;

    uniform_int_distribution<int> space_dist(o.min_spacing, o.max_spacing);
    discrete_distribution<char> type_dist(begin(o.type_weights), end(o.type_weights));

    if(s.no_overlap) assert(o.min_spacing > 0 && o.zero_space_chance == 0);

    Problem p;
    for(int i = 0; i < 3; ++i) {
        assert(o.max_travel_time[i] <= s.max_travel_times[i]);
        p.travel_times[i] = uniform_int_distribution<int>(o.min_travel_time[i], o.max_travel_time[i])(r);
    }
    sort(rbegin(p.travel_times), rend(p.travel_times));
    if(s.no_Ex) p.travel_times[Ex] = 1;
    int t = 0;
    for(int i = 0; i < n; ++i) {
        int dt = space_dist(r);
        if(prob_dist(r) < o.zero_space_chance) dt = 0;
        t += dt;
        if(t > s.max_arrival) break;
        Tp tp = (Tp)type_dist(r);
        p.trains.push_back({t, tp});
    }
    p.n = p.trains.size();
    return p;
}

struct OptionsTL {
    double space_weights[5];
    bool exact_n = false;
    bool type_weights_rand = true;
    double type_weights[3] = {0};
};
Problem gen_tiny_limited(rng r, Subtask s, OptionsTL o) {
    uniform_real_distribution<double> prob_dist(0, 1);

    int n = s.max_trains;
    if(!o.exact_n) n *= (9 + prob_dist(r))/10;

    discrete_distribution<int> space_dist(begin(o.space_weights), end(o.space_weights));

    if(o.type_weights_rand) {
        for(int i = 0; i < 2; ++i) {
            o.type_weights[i] = prob_dist(r);
        }
    }

    discrete_distribution<char> type_dist(begin(o.type_weights), end(o.type_weights));

    Problem p;
    p.travel_times[0] = 4;
    p.travel_times[1] = 1;
    p.travel_times[2] = 1;

    sort(rbegin(p.travel_times), rend(p.travel_times));
    if(s.no_Ex) p.travel_times[Ex] = 1;
    int t = 0;
    for(int i = 0; i < n; ++i) {
        int dt = space_dist(r);
        t += dt;
        if(t > s.max_arrival) break;
        Tp tp = (Tp)type_dist(r);
        p.trains.push_back({t, tp});
    }
    p.n = p.trains.size();
    return p;
}

array<Subtask, 8> subtasks = {{
    {.name = "max-8", .max_trains = 8},
    {.name = "max-14", .max_trains = 14},
    {.name = "max-40", .max_trains = 40},
    {.name = "tiny-values-limited", .max_trains = 100, .max_arrival = 1000, .max_travel_times = {4, 1, 1}, .no_Ex = true, .no_overlap = true},
    {.name = "small-values", .max_trains = 100, .max_arrival = 1000, .max_travel_times = {100, 100, 100}},
    {.name = "no-ex", .max_trains = 100, .no_Ex = true},
    {.name = "max-100", .max_trains = 100},
    {.name = "full", .max_trains = 400},
}};

int main(int argc, char *argv[]) {
    invocation invoc = init(argc, argv);
    auto builder = invoc.testcases();

    for (Subtask s : subtasks) {
        auto subtask_builder = builder.group(s.name, s);

        {
            int lim_rep = 2;
            if(s.name == "tiny-values-limited"s) lim_rep = 6;
            subtask_builder.repeated(lim_rep, "tiny-limited", gen, Options{
                    .min_spacing = 1,
                    .max_spacing = 2,
                    .min_travel_time = {3, 1, 1},
                    .max_travel_time = {4, 1, 1},
                    .type_weights_rand = false,
                    .type_weights = {1, 5},
                });
        }

        if(s.name == "tiny-values-limited"s) {
            subtask_builder.single("spaced", gen, Options{
                    .min_spacing = 1,
                    .max_spacing = 6,
                    .min_travel_time = {3, 1, 1},
                    .max_travel_time = {4, 1, 1},
                    .type_weights_rand = false,
                    .type_weights = {1, 5},
                });
            for(int w1 = 70; w1 <= 100; w1 += 3) {
                subtask_builder.single("space-weight-" + std::to_string(w1), gen_tiny_limited, OptionsTL{
                        .space_weights = {0, (0.01*w1), 1-0.01*w1, 0.05, 0},
                    });
            }
        }

        if(s.max_travel_times[0] >= 10) {
            for(int sp : {6, 8, 12}) {
                subtask_builder.single("small-" + to_string(sp), gen, Options{
                        .min_spacing = 1,
                        .max_spacing = sp,
                        .min_travel_time = {5, 3, 1},
                        .max_travel_time = {10, 5, 3},
                        .exact_n = true,
                    });
            }
        }

        if(s.max_travel_times[0] >= 100) {
            for(int sp : {14, 18, 20, 22, 40}) {
                subtask_builder.single("medium-" + to_string(sp), gen, Options{
                        .min_spacing = 1,
                        .max_spacing = sp,
                        .min_travel_time = {66, 33, 1},
                        .max_travel_time = {100, 66, 33},
                    });
            }
            for(int sp : {16, 20, 22, 26}) {
                subtask_builder.single("more-zeroes-medium-" + to_string(sp), gen, Options{
                        .min_spacing = 1,
                        .max_spacing = sp,
                        .zero_space_chance = 0.3,
                        .min_travel_time = {66, 33, 1},
                        .max_travel_time = {100, 66, 33},
                    });
            }
            for(int sp : {20, 28, 36}) {
                subtask_builder.single("many-zeroes-medium-" + to_string(sp), gen, Options{
                        .min_spacing = 1,
                        .max_spacing = sp,
                        .zero_space_chance = 0.6,
                        .min_travel_time = {66, 33, 1},
                        .max_travel_time = {100, 66, 33},
                    });
            }
            /* for(int sp = 4; sp <= 100; sp *= 1.8) {*/
            /*     subtask_builder.single("wild-travel-times-medium-" + to_string(sp), gen, Options{*/
            /*             .min_spacing = 0,*/
            /*             .max_spacing = sp,*/
            /*             .min_travel_time = {1, 1, 1},*/
            /*             .max_travel_time = {100, 100, 100},*/
            /*         });*/
            /* }*/
        }
        if(s.max_travel_times[0] >= 100000) {
            for(int sp = 3000; sp <= 2*s.max_travel_times[0]; sp *= 2) {
                subtask_builder.single("large-" + to_string(sp), gen, Options{
                        .min_spacing = 1,
                        .max_spacing = sp,
                        .zero_space_chance = 0.1,
                        .min_travel_time = {1, 1, 1},
                        .max_travel_time = {100000, 100000, 100000},
                    });
            }
        }
    }

    invoc.finish();
}
