#pragma once

#include <cstdint>
#include <exception>
#include <functional>
#include <iostream>
#include <random>
#include <sstream>
#include <string>
#include <tuple>
#include <type_traits>

#pragma GCC diagnostic ignored "-Wnon-template-friend"

typedef std::mt19937_64 rng;

template <typename Group>
class builder;
class root_input_group;

class invocation {
    template <typename Group>
    friend class builder;

    char const *testcase;
    char const *seed;

    bool testcase_matched = false;

    bool should_generate() const {
        return testcase != nullptr;
    }

    void report_match() {
        if (testcase_matched)
            throw std::runtime_error("There are multiple testcases with the given name.");
        testcase_matched = true;
    }

  public:
    invocation(char const *testcase = nullptr, char const *seed = nullptr) : testcase(testcase), seed(seed) {}

    void finish() {
        if (should_generate() && !testcase_matched)
            throw std::runtime_error("The requested testcase was not found.");
    }

    builder<root_input_group> testcases();
};

class root_input_group {
    template <typename Parent, typename Name, typename... Args>
    friend class input_group;
    template <typename Group>
    friend class builder;

    template <typename Function>
    auto bind(Function f) {
        return f;
    }

    template <typename Stream>
    void write_name(Stream &o) {}
};

template <typename Parent, typename Name, typename... Args>
class input_group {
    template <typename OtherParent, typename OtherName, typename... OtherArgs>
    friend class input_group;
    template <typename Group>
    friend class builder;

    Parent &parent;
    Name name;
    std::tuple<Args...> args;

    input_group(Parent &parent, Name name, Args... args) : parent(parent), name(name), args(args...) {}

    template <typename Function>
    auto bind(Function f) {
        auto parent_f = parent.bind(f);
        return std::apply(std::bind_front<decltype(parent_f), Args...>,
                          std::tuple_cat(std::make_tuple(parent_f), args));
    }

    template <typename Stream>
    void write_name(Stream &o) {
        parent.write_name(o);
        o << name << '-';
    }
};

template <typename Group>
class builder {
    friend class invocation;

    template <typename OtherGroup>
    friend class builder;

    Group target_group;
    invocation &invoc;

    builder(Group group, invocation &invoc) : target_group(group), invoc(invoc) {}

    void declare(std::string &name, std::uint32_t repeat, bool seeded) {
        std::cout << name
                  << " repeat=" << repeat
                  << " seeded=" << (seeded ? "true" : "false")
                  << '\n';
    }

    template <typename Name>
    std::string full_name(Name name) {
        std::ostringstream result;
        target_group.write_name(result);
        result << name;
        return result.str();
    }

    template <typename Function>
    auto bind(Function function) {
        return target_group.bind(function);
    }

    template <typename Function>
    auto bind_seed(Function function, std::uint_fast64_t seed) {
        rng r(seed);
        return bind(std::bind_front(function, r));
    }

  public:
    template <typename Function, typename Name, typename... Args>
    typename std::enable_if<std::is_invocable<decltype(std::declval<builder>().bind(std::declval<Function>())), Args...>::value>::type
    repeated(std::uint32_t count, Name name, Function function, Args... args) {
        std::string name_str = full_name(name);

        if (!invoc.should_generate()) {
            declare(name_str, count, false);
        } else {
            if (name_str == invoc.testcase) {
                invoc.report_match();
                auto result = bind(function)(args...);
                std::cout << result;
            }
        }
    }

    template <typename Function, typename Name, typename... Args>
    typename std::enable_if<std::is_invocable<decltype(std::declval<builder>().bind_seed(std::declval<Function>(), 0)), Args...>::value>::type
    repeated(std::uint32_t count, Name name, Function function, Args... args) {
        std::string name_str = full_name(name);

        if (!invoc.should_generate()) {
            declare(name_str, count, true);
        } else {
            if (name_str == invoc.testcase) {
                invoc.report_match();

                if (invoc.seed == nullptr)
                    throw std::runtime_error("This testcase is seeded, but no seed was given.");

                std::uint_fast64_t seed;
                std::istringstream seed_stream(invoc.seed);
                seed_stream.exceptions(std::ios_base::badbit | std::ios_base::failbit);
                seed_stream >> std::hex >> seed;

                auto result = bind_seed(function, seed)(args...);
                std::cout << result;
            }
        }
    }

    template <typename Function, typename Name, typename... Args>
    void single(Name name, Function function, Args... args) {
        repeated(1, name, function, args...);
    }

    template <typename Name, typename... Args>
    builder<input_group<Group, Name, Args...>> group(Name name, Args... args) {
        return builder<input_group<Group, Name, Args...>>(input_group(target_group, name, args...), invoc);
    }
};

inline builder<root_input_group> invocation::testcases() {
    return builder(root_input_group(), *this);
}

inline auto init(int argc, char *argv[]) {
    std::cout.exceptions(std::ios_base::badbit | std::ios_base::failbit);

    if (argc == 1)
        return invocation();
    else if (argc == 2)
        return invocation(argv[1]);
    else if (argc == 3)
        return invocation(argv[1], argv[2]);

    char const *name = (argc >= 1) ? argv[0] : "generate";

    std::ostringstream message;
    message << "Usage: " << name << " [testcase] [seed]";
    throw std::runtime_error(message.str());
}

